@@ -5,25 +5,24 @@
 # License, v. 2.0. If a copy of the MPL was not distributed with this file,
 # You can obtain one at http://mozilla.org/MPL/2.0/.
 #
-# Author: Kyle Lahnakoski (kyle@lahnakoski.com)
+# Contact: Kyle Lahnakoski (kyle@lahnakoski.com)
 #
 import base64
 import io
+import os
 import re
 import shutil
 from datetime import datetime
 from mimetypes import MimeTypes
-from tempfile import mkdtemp, NamedTemporaryFile
+from tempfile import NamedTemporaryFile, mkdtemp
 
-import os
+from mo_dots import Null, coalesce, get_module, is_list
+from mo_files import mimetype
+from mo_files.url import URL
+from mo_future import PY3, binary_type, text, is_text
+from mo_logs import Except, Log
+from mo_logs.exceptions import get_stacktrace
 
-from mo_future import text_type, binary_type
-from mo_dots import get_module, coalesce, Null
-from mo_logs import Log, Except
-from mo_logs.exceptions import extract_stack
-from mo_threads import Thread, Till
-
-mime = MimeTypes()
 
 class File(object):
     """
@@ -42,35 +41,40 @@ class File(object):
         """
         YOU MAY SET filename TO {"path":p, "key":k} FOR CRYPTO FILES
         """
-        self._mime_type = mime_type
-        if filename == None:
-            Log.error(u"File must be given a filename")
-        elif isinstance(filename, File):
+        if isinstance(filename, File):
             return
-        elif isinstance(filename, (binary_type, text_type)):
-            self.key = None
-            if filename==".":
-                self._filename = ""
-            elif filename.startswith("~"):
-                home_path = os.path.expanduser("~")
-                if os.sep == "\\":
-                    home_path = home_path.replace(os.sep, "/")
-                if home_path.endswith("/"):
-                    home_path = home_path[:-1]
-                filename = home_path + filename[1::]
-            self._filename = filename.replace(os.sep, "/")  # USE UNIX STANDARD
+
+        self._mime_type = mime_type
+
+        if isinstance(filename, (binary_type, text)):
+            try:
+                self.key = None
+                if filename==".":
+                    self._filename = ""
+                elif filename.startswith("~"):
+                    home_path = os.path.expanduser("~")
+                    if os.sep == "\\":
+                        home_path = home_path.replace(os.sep, "/")
+                    if home_path.endswith("/"):
+                        home_path = home_path[:-1]
+                    filename = home_path + filename[1::]
+                self._filename = filename.replace(os.sep, "/")  # USE UNIX STANDARD
+            except Exception as e:
+                Log.error(u"can not load {{file}}", file=filename, cause=e)
         else:
-            self.key = base642bytearray(filename.key)
-            self._filename = "/".join(filename.path.split(os.sep))  # USE UNIX STANDARD
+            try:
+                self.key = base642bytearray(filename.key)
+                self._filename = "/".join(filename.path.split(os.sep))  # USE UNIX STANDARD
+            except Exception as e:
+                Log.error(u"can not load {{file}}", file=filename.path, cause=e)
 
         while self._filename.find(".../") >= 0:
             # LET ... REFER TO GRANDPARENT, .... REFER TO GREAT-GRAND-PARENT, etc...
             self._filename = self._filename.replace(".../", "../../")
         self.buffering = buffering
 
-
         if suffix:
-            self._filename = File.add_suffix(self._filename, suffix)
+            self._filename = add_suffix(self._filename, suffix)
 
     @classmethod
     def new_instance(cls, *path):
@@ -110,17 +114,11 @@ class File(object):
             else:
                 return os.path.abspath(self._filename)
 
-    @staticmethod
-    def add_suffix(filename, suffix):
+    def add_suffix(self, suffix):
         """
         ADD suffix TO THE filename (NOT INCLUDING THE FILE EXTENSION)
         """
-        path = filename.split("/")
-        parts = path[-1].split(".")
-        i = max(len(parts) - 2, 0)
-        parts[i] = parts[i] + suffix
-        path[-1] = ".".join(parts)
-        return "/".join(path)
+        return File(add_suffix(self._filename, suffix))
 
     @property
     def extension(self):
@@ -146,8 +144,9 @@ class File(object):
             elif self.abspath.endswith(".css"):
                 self._mime_type = "text/css"
             elif self.abspath.endswith(".json"):
-                self._mime_type = "application/json"
+                self._mime_type = mimetype.JSON
             else:
+                mime = MimeTypes()
                 self._mime_type, _ = mime.guess_type(self.abspath)
                 if not self._mime_type:
                     self._mime_type = "application/binary"
@@ -183,6 +182,12 @@ class File(object):
         path[-1] = ".".join(parts)
         return File("/".join(path))
 
+    def add_extension(self, ext):
+        """
+        RETURN NEW FILE WITH EXTENSION ADDED (OLD EXTENSION IS A SUFFIX)
+        """
+        return File(self._filename + "." + text(ext))
+
     def set_name(self, name):
         """
         RETURN NEW FILE WITH GIVEN EXTENSION
@@ -225,15 +230,16 @@ class File(object):
             for num, zip_name in enumerate(zipped.namelist()):
                 return zipped.open(zip_name).read().decode(encoding)
 
-
     def read_lines(self, encoding="utf8"):
         with open(self._filename, "rb") as f:
             for line in f:
                 yield line.decode(encoding).rstrip()
 
     def read_json(self, encoding="utf8", flexible=True, leaves=True):
+        from mo_json import json2value
+
         content = self.read(encoding=encoding)
-        value = get_module(u"mo_json").json2value(content, flexible=flexible, leaves=leaves)
+        value = json2value(content, flexible=flexible, leaves=leaves)
         abspath = self.abspath
         if os.sep == "\\":
             abspath = "/" + abspath.replace(os.sep, "/")
@@ -267,21 +273,21 @@ class File(object):
         if not self.parent.exists:
             self.parent.create()
         with open(self._filename, "wb") as f:
-            if isinstance(data, list) and self.key:
+            if is_list(data) and self.key:
                 Log.error(u"list of data and keys are not supported, encrypt before sending to file")
 
-            if isinstance(data, list):
+            if is_list(data):
                 pass
-            elif isinstance(data, (binary_type, text_type)):
+            elif isinstance(data, (binary_type, text)):
                 data=[data]
             elif hasattr(data, "__iter__"):
                 pass
 
             for d in data:
-                if not isinstance(d, text_type):
+                if not is_text(d):
                     Log.error(u"Expecting unicode data only")
                 if self.key:
-                    from mo_math.crypto import encrypt
+                    from mo_math.aes_crypto import encrypt
                     f.write(encrypt(d, self.key).encode("utf8"))
                 else:
                     f.write(d.encode("utf8"))
@@ -312,7 +318,7 @@ class File(object):
         if not self.parent.exists:
             self.parent.create()
         with open(self._filename, "ab") as output_file:
-            if not isinstance(content, text_type):
+            if not is_text(content):
                 Log.error(u"expecting to write unicode only")
             output_file.write(content.encode(encoding))
             output_file.write(b"\n")
@@ -329,7 +335,7 @@ class File(object):
                 self.parent.create()
             with open(self._filename, "ab") as output_file:
                 for c in content:
-                    if isinstance(c, str):
+                    if not isinstance(c, text):
                         Log.error(u"expecting to write unicode only")
 
                     output_file.write(c.encode("utf8"))
@@ -419,9 +425,15 @@ class File(object):
     def copy(cls, from_, to_):
         _copy(File(from_), File(to_))
 
+    def __data__(self):
+        return self._filename
+
     def __unicode__(self):
         return self.abspath
 
+    def __str__(self):
+        return self.abspath
+
 
 class TempDirectory(File):
     """
@@ -429,7 +441,7 @@ class TempDirectory(File):
     WILL BE DELETED WHEN EXITED
     """
     def __new__(cls):
-        return File.__new__(cls, None)
+        return object.__new__(cls)
 
     def __init__(self):
         File.__init__(self, mkdtemp())
@@ -438,7 +450,9 @@ class TempDirectory(File):
         return self
 
     def __exit__(self, exc_type, exc_val, exc_tb):
-        Thread.run("delete dir " + self.name, delete_daemon, file=self, caller_stack=extract_stack(1))
+        from mo_threads import Thread
+
+        Thread.run("delete dir " + self.name, delete_daemon, file=self, caller_stack=get_stacktrace(1)).release()
 
 
 class TempFile(File):
@@ -449,7 +463,9 @@ class TempFile(File):
     def __new__(cls, *args, **kwargs):
         return object.__new__(cls)
 
-    def __init__(self):
+    def __init__(self, filename=None):
+        if isinstance(filename, File):
+            return
         self.temp = NamedTemporaryFile(delete=False)
         self.temp.close()
         File.__init__(self, self.temp.name)
@@ -458,8 +474,9 @@ class TempFile(File):
         return self
 
     def __exit__(self, exc_type, exc_val, exc_tb):
-        Thread.run("delete file " + self.name, delete_daemon, file=self, caller_stack=extract_stack(1))
+        from mo_threads import Thread
 
+        Thread.run("delete file " + self.name, delete_daemon, file=self, caller_stack=get_stacktrace(1)).release()
 
 def _copy(from_, to_):
     if from_.is_directory():
@@ -469,11 +486,18 @@ def _copy(from_, to_):
         File.new_instance(to_).write_bytes(File.new_instance(from_).read_bytes())
 
 
-def base642bytearray(value):
-    if value == None:
-        return bytearray("")
-    else:
-        return bytearray(base64.b64decode(value))
+if PY3:
+    def base642bytearray(value):
+        if value == None:
+            return bytearray(b"")
+        else:
+            return bytearray(base64.b64decode(value))
+else:
+    def base642bytearray(value):
+        if value == None:
+            return bytearray(b"")
+        else:
+            return bytearray(base64.b64decode(value))
 
 
 def datetime2string(value, format="%Y-%m-%d %H:%M:%S"):
@@ -539,6 +563,8 @@ def join_path(*path):
 
 def delete_daemon(file, caller_stack, please_stop):
     # WINDOWS WILL HANG ONTO A FILE FOR A BIT AFTER WE CLOSED IT
+    from mo_threads import Till
+
     while not please_stop:
         try:
             file.delete()
@@ -549,3 +575,16 @@ def delete_daemon(file, caller_stack, please_stop):
 
             Log.warning(u"problem deleting file {{file}}", file=file.abspath, cause=e)
             (Till(seconds=10)|please_stop).wait()
+
+
+def add_suffix(filename, suffix):
+    """
+    ADD suffix TO THE filename (NOT INCLUDING THE FILE EXTENSION)
+    """
+    path = filename.split("/")
+    parts = path[-1].split(".")
+    i = max(len(parts) - 2, 0)
+    parts[i] = parts[i] + "." + text(suffix).strip(".")
+    path[-1] = ".".join(parts)
+    return File("/".join(path))
+
