@@ -1,7 +1,14 @@
-from collections import Mapping
+# encoding: utf-8
+#
+# This Source Code Form is subject to the terms of the Mozilla Public
+# License, v. 2.0. If a copy of the MPL was not distributed with this file,
+# You can obtain one at http://mozilla.org/MPL/2.0/.
+#
+# Contact: Kyle Lahnakoski (kyle@lahnakoski.com)
+#
 
-from mo_dots import wrap, Data
-from mo_future import urlparse, text_type, PY2, unichr
+from mo_dots import Data, Null, coalesce, is_data, is_list, wrap
+from mo_future import PY2, is_text, text, unichr, urlparse, is_binary
 from mo_logs import Log
 
 
@@ -12,14 +19,22 @@ class URL(object):
     [1] https://docs.python.org/3/library/urllib.parse.html
     """
 
-    def __init__(self, value):
+    def __new__(cls, value, *args, **kwargs):
+        if isinstance(value, URL):
+            return value
+        else:
+            return object.__new__(cls)
+
+    def __init__(self, value, port=None, path=None, query=None, fragment=None):
+        if isinstance(value, URL):
+            return
         try:
             self.scheme = None
             self.host = None
-            self.port = None
-            self.path = ""
-            self.query = ""
-            self.fragment = ""
+            self.port = port
+            self.path = path
+            self.query = query
+            self.fragment = fragment
 
             if value == None:
                 return
@@ -33,13 +48,13 @@ class URL(object):
             else:
                 output = urlparse(value)
                 self.scheme = output.scheme
-                self.port = output.port
+                self.port = coalesce(port, output.port)
                 self.host = output.netloc.split(":")[0]
-                self.path = output.path
-                self.query = wrap(url_param2value(output.query))
-                self.fragment = output.fragment
+                self.path = coalesce(path, output.path)
+                self.query = coalesce(query, wrap(url_param2value(output.query)))
+                self.fragment = coalesce(fragment, output.fragment)
         except Exception as e:
-            Log.error("problem parsing {{value}} to URL", value=value, cause=e)
+            Log.error(u"problem parsing {{value}} to URL", value=value, cause=e)
 
     def __nonzero__(self):
         if self.scheme or self.host or self.port or self.path or self.query or self.fragment:
@@ -52,7 +67,7 @@ class URL(object):
         return False
 
     def __truediv__(self, other):
-        if not isinstance(other, text_type):
+        if not is_text(other):
             Log.error(u"Expecting text path")
         output = self.__copy__()
         output.path = output.path.rstrip('/') + "/" + other.lstrip('/')
@@ -71,6 +86,11 @@ class URL(object):
         output.fragment = self.fragment
         return output
 
+    def decode(self, encoding=''):
+        return text(self).decode(encoding)
+
+    def __data__(self):
+        return str(self)
 
     def __str__(self):
         url = ""
@@ -81,14 +101,14 @@ class URL(object):
         if self.port:
             url = url + ":" + str(self.port)
         if self.path:
-            if self.path[0] == text_type("/"):
+            if self.path[0] == text("/"):
                 url += str(self.path)
             else:
                 url += "/" + str(self.path)
         if self.query:
-            url = url + '?' + value2url_param(self.query)
+            url = url + "?" + value2url_param(self.query)
         if self.fragment:
-            url = url + '#' + value2url_param(self.fragment)
+            url = url + "#" + value2url_param(self.fragment)
         return url
 
 
@@ -96,6 +116,12 @@ def int2hex(value, size):
     return (("0" * size) + hex(value)[2:])[-size:]
 
 
+def hex2chr(hex):
+    try:
+        return unichr(int(hex, 16))
+    except Exception as e:
+        raise e
+
 if PY2:
     _map2url = {chr(i): chr(i) for i in range(32, 128)}
     for c in " {}<>;/?:@&=+$,":
@@ -103,11 +129,11 @@ if PY2:
     for i in range(128, 256):
         _map2url[chr(i)] = "%" + str(int2hex(i, 2))
 else:
-    _map2url = {unichr(i): unichr(i) for i in range(32, 128)}
-    for c in " {}<>;/?:@&=+$,":
-        _map2url[c] = "%" + int2hex(ord(c), 2)
+    _map2url = {i: unichr(i) for i in range(32, 128)}
+    for c in b" {}<>;/?:@&=+$,":
+        _map2url[c] = "%" + int2hex(c, 2)
     for i in range(128, 256):
-        _map2url[unichr(i)] = "%" + str(int2hex(i, 2))
+        _map2url[i] = "%" + str(int2hex(i, 2))
 
 
 names = ["path", "query", "fragment"]
@@ -131,8 +157,10 @@ def url_param2value(param):
     """
     CONVERT URL QUERY PARAMETERS INTO DICT
     """
-    if isinstance(param, text_type):
-        param = param.encode("ascii")
+    if param == None:
+        return Null
+    if param == None:
+        return Null
 
     def _decode(v):
         output = []
@@ -140,35 +168,37 @@ def url_param2value(param):
         while i < len(v):
             c = v[i]
             if c == "%":
-                d = (v[i + 1:i + 3]).decode("hex")
+                d = hex2chr(v[i + 1:i + 3])
                 output.append(d)
                 i += 3
             else:
                 output.append(c)
                 i += 1
 
-        output = (b"".join(output)).decode("latin1")
+        output = text("".join(output))
         try:
+            from mo_json import json2value
+
             return json2value(output)
         except Exception:
             pass
         return output
 
     query = Data()
-    for p in param.split(b'&'):
+    for p in param.split('&'):
         if not p:
             continue
-        if p.find(b"=") == -1:
+        if p.find("=") == -1:
             k = p
             v = True
         else:
-            k, v = p.split(b"=")
+            k, v = p.split("=")
             v = _decode(v)
 
         u = query.get(k)
         if u is None:
             query[k] = v
-        elif isinstance(u, list):
+        elif is_list(u):
             u += [v]
         else:
             query[k] = [u, v]
@@ -184,18 +214,20 @@ def value2url_param(value):
     if value == None:
         Log.error("Can not encode None into a URL")
 
-    if isinstance(value, Mapping):
+    if is_data(value):
+        from mo_json import value2json
+
         value_ = wrap(value)
-        output = b"&".join([
-            value2url_param(k) + b"=" + (value2url_param(v) if isinstance(v, text_type) else value2url_param(value2json(v)))
+        output = "&".join([
+            value2url_param(k) + "=" + (value2url_param(v) if is_text(v) else value2url_param(value2json(v)))
             for k, v in value_.leaves()
             ])
-    elif isinstance(value, text_type):
-        output = b"".join(_map2url[c] for c in value.encode('utf8'))
-    elif isinstance(value, str):
-        output = b"".join(_map2url[c] for c in value)
+    elif is_text(value):
+        output = "".join(_map2url[c] for c in value.encode('utf8'))
+    elif is_binary(value):
+        output = "".join(_map2url[c] for c in value)
     elif hasattr(value, "__iter__"):
-        output = b",".join(value2url_param(v) for v in value)
+        output = ",".join(value2url_param(v) for v in value)
     else:
         output = str(value)
     return output
